#include <stdio.h>
#include <stdlib.h>

#ifdef DEBUG
#define D if (1)
#else
#define D if (0)
#endif

#define NUM_LOOPS 8192
#define NUM_PER_LOOP 1024

struct ExtMemMoveStruct
{
    unsigned long int length;
    unsigned int source_handle;
    unsigned long int source_offset;
    unsigned int dest_handle;
    unsigned long int dest_offset;
};

void *
xcalloc(size_t nmemb, size_t size)
{
    void *ptr;

    ptr = calloc(nmemb, size);
    if (ptr == NULL)
    {
        perror("xcalloc");
        abort();
    }
    return ptr;
}

char
xms_driver_installed(void)
{
    unsigned char res;

    asm {
        mov ax, 0x4300
        int 0x2F
        mov res, al
    };
    return res == 0x80;
}

unsigned long int
xms_get_control_function(void)
{
    unsigned long int res;

    asm {
        mov  ax, 0x4310
        int  0x2F
        mov  word ptr [res], bx
        mov  word ptr [res + 2], es
    };
    return res;
}

void
xms_get_version(
    unsigned long int xms_control,
    unsigned int *version,
    unsigned int *internal,
    unsigned int *hma_present
)
{
    unsigned int v, i, h;

    asm {
        mov   ah, 0x00
        call  [xms_control]
        mov   v, ax
        mov   i, bx
        mov   h, dx
    };
    *version = v;
    *internal = i;
    *hma_present = h;
}

void
xms_query_free_extended(
    unsigned long int xms_control,
    unsigned int *free_largest_k,
    unsigned int *free_total_k
)
{
    unsigned int l, t;

    asm {
        mov   ah, 0x08
        call  [xms_control]
        mov   l, ax
        mov   t, dx
    };
    *free_largest_k = l;
    *free_total_k = t;
}

char
xms_alloc_extended(
    unsigned long int xms_control,
    unsigned int request_k,
    unsigned int *handle
)
{
    unsigned int success, h;

    asm {
        mov   ah, 0x09
        mov   dx, request_k
        call  [xms_control]
        mov   success, ax
        mov   h, dx
    };
    *handle = h;
    return success == 0x0001;
}

char
xms_move_extended(
    unsigned long int xms_control,
    struct ExtMemMoveStruct *m
)
{
    unsigned int success;

    asm {
        mov   ah, 0x0B
        lds   si, m
        call  [xms_control]
        mov   success, ax
    };
    return success == 0x0001;
}

char
xms_free_extended(unsigned long int xms_control, unsigned int handle)
{
    unsigned int success;

    asm {
        mov   ah, 0x0A
        mov   dx, handle
        call  [xms_control]
        mov   success, ax
    };
    return success == 0x0001;
}

int
main()
{
    unsigned long int far *mydata;
    unsigned long int loop, expected;
    unsigned int idx, ext_mem_size_kb;
    unsigned long int xms_control;
    unsigned int vers = 0, internal = 0, hma = 0;
    unsigned int free_largest_k = 0, free_total_k = 0;
    unsigned int xms_handle = 0;
    struct ExtMemMoveStruct move_struct;

    if (!xms_driver_installed())
    {
        fprintf(stderr, "No XMS driver installed.\n");
        return 1;
    }

    /* This is mandatory: We need to know where the XMS control function
     * lives in memory. */
    xms_control = xms_get_control_function();
    printf("xms_control %lu\n", xms_control);

    /* Informational. */
    xms_get_version(xms_control, &vers, &internal, &hma);
    printf("XMS version: (%02X, %02X, %02X)\n", vers, internal, hma);

    /* Informational. */
    xms_query_free_extended(xms_control, &free_largest_k, &free_total_k);
    printf(
        "Extended mem: Largest free block %u kB, free total %u kB\n",
        free_largest_k, free_total_k
    );

    /* This will allocate a large block of extended memory. The only way
     * for us to address it is by using its handle. We don't (and can't)
     * get a pointer directly.
     *
     * Size is kB, so we don't multiply it with NUM_PER_LOOP here. */
    ext_mem_size_kb = NUM_LOOPS * sizeof (unsigned long int);
    printf("Allocating %u kB of extended memory ...\n", ext_mem_size_kb);
    if (!xms_alloc_extended(
        xms_control,
        ext_mem_size_kb,
        &xms_handle
    ))
    {
        fprintf(stderr, "xms_alloc_extended() failed\n");
        return 1;
    }

    printf("Copying test data around ...\n");

    for (loop = 0; loop < NUM_LOOPS; loop++)
    {
        D printf("Creating data on heap (conventional memory), loop %d \n", loop);
        mydata = xcalloc(NUM_PER_LOOP, sizeof (unsigned long int));
        for (idx = 0; idx < NUM_PER_LOOP; idx++)
            mydata[idx] = loop * NUM_PER_LOOP + idx;

        D printf("Copying data from heap to extended memory ...\n");
        move_struct.length = NUM_PER_LOOP * sizeof (unsigned long int);
        move_struct.source_handle = 0;  /* source is conventional mem */
        move_struct.source_offset = (unsigned long int)mydata;
        move_struct.dest_handle = xms_handle;
        move_struct.dest_offset = loop * NUM_PER_LOOP * sizeof (unsigned long int);
        if (!xms_move_extended(xms_control, &move_struct))
        {
            fprintf(stderr, "xms_move_extended() failed (to extmem)\n");
            return 1;
        }

        D printf("Freeing pointer to memory on heap ...\n");
        free(mydata);
    }

    for (loop = 0; loop < NUM_LOOPS; loop++)
    {
        D printf("Allocating new memory on heap, loop %d ...\n", loop);
        mydata = xcalloc(NUM_PER_LOOP, sizeof (unsigned long int));

        D printf("Moving data back from extended memory to heap ...\n");
        move_struct.length = NUM_PER_LOOP * sizeof (unsigned long int);
        move_struct.source_handle = xms_handle;
        move_struct.source_offset = loop * NUM_PER_LOOP * sizeof (unsigned long int);
        move_struct.dest_handle = 0;  /* dest is conventional mem */
        move_struct.dest_offset = (unsigned long int)mydata;
        if (!xms_move_extended(xms_control, &move_struct))
        {
            fprintf(stderr, "xms_move_extended() failed (from extmem)\n");
            return 1;
        }

        D printf("Comparing data with expected values ...\n");
        for (idx = 0; idx < NUM_PER_LOOP; idx++)
        {
            expected = loop * NUM_PER_LOOP + idx;
            D printf("Loop %lu, idx %u: %lu\n", loop, idx, mydata[idx]);
            if (mydata[idx] != expected)
            {
                fprintf(
                    stderr,
                    "Fail in loop %lu, idx %u: %lu vs. %lu\n",
                    loop, idx,
                    mydata[idx], expected
                );
                return 1;
            }
        }

        free(mydata);
    }

    /* Comment out this block to see how the available extended memory
     * decreases (see output of MEM on the command line) even after this
     * program exits. */
    if (!xms_free_extended(xms_control, xms_handle))
    {
        fprintf(stderr, "Failed to free extended memory\n");
        return 1;
    }

    printf("All good.\n");

    return 0;
}
